import java.io.BufferedReader;
import java.io.DataInputStream;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.maven.plugin.AbstractMojo;
import org.apache.maven.plugin.MojoExecutionException;

/**
 * Suck the dependencies from another project
 *
 * @goal absorb
 */
public class AssimilateMojo extends AbstractMojo{
	/**
	 * The specific Maven command options and commands needed to be executed to build the current project.
	 * For example: 'clean install' or '-DskipTests clean package'
	 * This is an OPTIONAL parameter.
	 * 
	 * @parameter default-value="clean install"
	 */
	private String mvnCommand;
	/**
	 * The location of Maven's Local Repository if it is not the default location, which is usually ${user.home}/.m2/repo
	 * This is an OPTIONAL parameter.
	 *
	 * @parameter 
	 */
	private String localRepository;
	/**
     * Whether Managed Dependencies in the External Project POMs 
     * have been defined at the bottom of the pom file (beneath all the other
     * configurations such as default build section and other profiles configurations.
     * This is an OPTIONAL parameter.
     *
     * @parameter default-value="false"
     */
    private String bottom;
	/**
	 * The List of Dependencies listed, each dependency representing a single prebuilt Maven project
	 * that exists in the local repository.
	 * This is a MANDATORY parameter.
	 *
	 * @parameter 
	 * @required
	 */
	private String[] dependencies;
	
	//Non-Mojo Parameters
	private String pathSeparator;
	private String mavenExecutionPrefix;
	
	public void execute() throws MojoExecutionException{
		
		validateBottomParameter();
		validateDependenciesParameter();
		resolveOperatingSystemQuirks();
		resolveLocalRepositoryLocation();
		
		String currentpompath = System.getProperty("user.dir")+getPathSeparator()+"pom.xml";
		getLog().info("assimilate::> current logged-on user directory:-> "+System.getProperty("user.home"));
		getLog().info("assimilate::> path to current project maven pom:-> "+currentpompath);
		String currentpom = readPom(currentpompath, new String());
		getLog().info("assimilate::> found "+dependencies.length+" external dependency(s) to assimilate");
		getLog().info("\n\n\n");
		String pomwithimporteddependencies = currentpom;
		for(String dependency: dependencies) {
			getLog().info("assimilate::> dealing with dependency:-> "+dependency);
			String path = localRepository+getPathSeparator()+".m2"+getPathSeparator()+"repository"+getPathSeparator()+dependencies[0].split(":")[0]+getPathSeparator()+dependencies[0].split(":")[1]+getPathSeparator()+
			               dependencies[0].split(":")[2]+getPathSeparator()+dependencies[0].split(":")[1]+"-"+dependencies[0].split(":")[2]+".pom";
			getLog().info("assimilate::> path to pom in local repository:-> "+path);
			String pomFromAgent=readPom(path, new String());
			
		    String pomWithInfusedDependencies = getDependencyStringGivenPomAndLocation(pomFromAgent, bottom);
		    getLog().info("assimilate::> managed-dependency subsection extracted from external pom: ");
		    getLog().info(pomWithInfusedDependencies.toString());
		    pomwithimporteddependencies = pomwithimporteddependencies.replaceFirst("</dependencies>", pomWithInfusedDependencies);
		    getLog().info("assimilate::> done with "+dependency);
		    getLog().info("\n\n");
		}
			
		String commentouthead = pomwithimporteddependencies.replaceAll("<plugin>\\s*<groupId>net\\.sf\\.assimilate", "<!--<plugin><groupId>net.sf.assimilate");
		String finalpom = commentouthead.replaceAll("absorb</goal>\\s*</goals>\\s*</execution>\\s*</executions>\\s*</plugin>", "absorb</goal></goals></execution></executions></plugin>-->");
					
		try {
			String pathToFinalPom = System.getProperty("user.dir")+getPathSeparator()+"assimilated_pom.xml";
			getLog().info("assimilate::> path to assimilated pom file:-> "+pathToFinalPom);
			FileOutputStream fos = new FileOutputStream(pathToFinalPom);
			fos.write(finalpom.getBytes());
			fos.close();
			getLog().info("assimilate::> successfully wrote final assimilated pom to: "+pathToFinalPom);
			
			
			String finalCommand = getMavenExecutionPrefix()+" -f "+System.getProperty("user.dir")+getPathSeparator()+"assimilated_pom.xml "+mvnCommand;
			getLog().info("assimilate::> command to be executed on newly assimilated pom: "+finalCommand);
			getLog().info("assimilate::> executing....");
			Process p = Runtime.getRuntime().exec(finalCommand);
			BufferedReader in = new BufferedReader(new InputStreamReader(p.getInputStream()));  
		    String theline = null;  
		     while ((theline = in.readLine()) != null) {  
		         System.out.println(theline);  
		     }  
		 } catch (Exception e) {
			 e.printStackTrace();
			 }  
		 getLog().info("assimilate::> assimilation complete.");
	}

	private String readPom(String currentpompath, String currentpom) {
		String line;
		try {
			
		    BufferedReader br = new BufferedReader(new InputStreamReader(new DataInputStream(new FileInputStream(currentpompath))));
		    while ((line = br.readLine()) != null)   {
		    	currentpom = currentpom+line;
		      
		    }
		    br.close();
		}
		catch(Exception e) {
		}
		return currentpom;
	}

	private void resolveLocalRepositoryLocation() {
		if(localRepository == null || localRepository.equals("")) {
			localRepository = System.getProperty("user.home");
		}
	}

	private void resolveOperatingSystemQuirks() {
		String pathSep = System.getProperty("os.name");
		if(pathSep.contains("Windows") || pathSep.contains("windows")) {
			setPathSeparator("\\");
			setMavenExecutionPrefix("cmd /C mvn");
		}
		else {
			setPathSeparator("/");
			setMavenExecutionPrefix("mvn");
		}
	}

	private void validateDependenciesParameter() {
		if(!verifyDependencyElementsExistence(dependencies)) {
			getLog().info("Mandatory '<dependencies>' Parameter in net.sf.assimilate:assimilate plugin contains a <dependency> element with invalid/missing content {groupId, artifactId or version missing in format: groupId:artifactId:version}.");
			System.exit(1);
		}
	}

	private void validateBottomParameter() {
		if(!bottom.equals("false") && !bottom.equalsIgnoreCase("true")) {
			getLog().info("Optional '<bottom>' Parameter Value in net.sf.assimilate:assimilate plugin not recognized! Use 'true' or 'false' (without quotes) to specify or leave out completely.");
			System.exit(1);
		}
	}

	protected String getDependencyStringGivenPomAndLocation(String fullpom, String bottom) {
		Pattern pattern = Pattern.compile("<dependency>.+?</dependencies>");
		Matcher matcher = pattern.matcher(fullpom);
		String dependenciesAsString = null;
		
		if("false".equals(bottom) || null == bottom){
		StringBuffer sb = new StringBuffer();
		 while (matcher.find()) {
			 sb.append(matcher.group());
			 break;
		 }
		 dependenciesAsString =  sb.toString();
		}
		
		else if("true".equalsIgnoreCase(bottom)){
			List<String> listOfDependencies = new ArrayList<String>();
			 while (matcher.find()) {
				 listOfDependencies.add(matcher.group());
			 }
		 dependenciesAsString =  listOfDependencies.get(listOfDependencies.size()-1);
			}
		
		return dependenciesAsString;
	}

	protected boolean verifyDependencyElementsExistence(String[] strings) {
		boolean allElementsPresent = true;
		for(String singleDependency:strings) {
		   String[] elements = singleDependency.split(":");
		    if(elements.length != 3) {
				allElementsPresent = false;
				break;
			}
			for(String element:elements) {
				if(element == null || element.equals("")) {
					allElementsPresent = false;
					break;
				}	
			}
		}
		return allElementsPresent;
	}

	private void setPathSeparator(String pathSeparator) {
		this.pathSeparator = pathSeparator;
	}

	private String getPathSeparator() {
		return pathSeparator;
	}

	private void setMavenExecutionPrefix(String mavenExecutionPrefix) {
		this.mavenExecutionPrefix = mavenExecutionPrefix;
	}

	private String getMavenExecutionPrefix() {
		return mavenExecutionPrefix;
	}
	
	
}